/***************************************************************************
 *  Copyright (C) 2011 by Resara LLC                                       *
 *  brendan@resara.com                                                     *
 *                                                                         *
 *  This program is free software; you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation; either version 2 of the License, or (at  *
 *  your option) any later version.                                        *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful, but    *
 *  WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU      *
 *  General Public License for more details.                               *
 *                                                                         *
 *  You should have received a copy of the GNU General Public License      *
 *  along with this program; if not, write to the                          *
 *  Free Software Foundation, Inc.,                                        *
 *  59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 *                                                                         *
 ***************************************************************************/


#include "dhcphosteditwidget.h"
#include <RdsEntityManager>
#include <QMessageBox>
#include <QDebug>
#include <RdsUtils>
#include <RdsDaemonManager>
#include <RdsFileManager>
#include <RdsClient>
#include <RdsDhcpValues>

#define IP_VALIDATOR QRegExp("^([0-9]{0,3}\\.){0,3}[0-9]{0,3}$")
#define IP_VALIDATOR_STRONG QRegExp("^(([0-9][0-9]?)|(1[0-9][0-9]?)|(2[0-4][0-9])|(25[0-5]))\\.(([0-9][0-9]?)|(1[0-9][0-9]?)|(2[0-4][0-9])|(25[0-5]))\\.(([0-9][0-9]?)|(1[0-9][0-9]?)|(2[0-4][0-9])|(25[0-5]))\\.(([0-9][0-9]?)|(1[0-9][0-9]?)|(2[0-4][0-9])|(25[0-5]))$")


DhcpHostEditWidget::DhcpHostEditWidget(RdsDhcpManager *manager, QWidget* parent)
		: RdsEntityWidget(manager, parent), Ui::DhcpHostEditWidget()
{
	setupUi(this);
	setApplyButton(ApplyButton);
	setDiscardButton(DiscardButton);

	_manager = manager;
	_configured = true;

	addWidget("name", Name, NoMulti, "^[^/]+$", "The name is not valid.");
	addWidget("ip", Ip, NoMulti, IP_VALIDATOR.pattern(), "The IP address is not valid.");
	addWidget("mac", Mac, NoMulti, "^[0-9a-fA-F:]*$", "The gateway must be a valid IP address.");
	addWidget("domainname", DomainName, None, "^[A-Za-z][A-Za-z0-9.\\-]*$", "The domain name is not valid.");
	addWidget("gateway", Gateway, None, IP_VALIDATOR.pattern(), "The gateway must be a valid IP address.");
	addWidget("dnsservers", DnsServers, "listString", SIGNAL(changed()));
	DnsServers->setSeperator(QRegExp(",[ ]*"), ", ");
	DnsServers->setErrorText("You must specify a valid IP address.");
	DnsServers->setWeakValidator(IP_VALIDATOR);
	DnsServers->setStrongValidator(IP_VALIDATOR_STRONG);

	QObject::connect(manager, SIGNAL(entityRenamed(QString, QString)), this, SLOT(entityRenamed(QString, QString)));
}

DhcpHostEditWidget::~DhcpHostEditWidget()
{
}

void DhcpHostEditWidget::entityRenamed(QString oldid, QString newid)
{
	if (oldid != input()) return;

	QStringList outlist = outputs();
	for (int i = 0; i < outlist.size(); i++)
	{
		if (outlist[i] == input()) outlist[i] = newid;
	}

	setInput(newid);
	setOutputs(outlist);

	//setField("name", newid);
}

ReturnValue DhcpHostEditWidget::getData()
{
	QVariantMap fields;

	if (input() == "") return(ReturnValue(1, "Invalid Folder"));

	ReturnValue err = true;

	ReturnValue ret = _manager->values(input());
	if (ret.isError()) err = ret;
	else
	{
		RdsDhcpValues values = ret;

		ret = values.value("fixed-address");
		if (ret.isError()) err = ret;
		fields["ip"] = ret;

		ret = values.value("hardware ethernet");
		if (ret.isError()) err = ret;
		fields["mac"] = ret;

		ret = values.option("domain-name");
		if (ret.isError()) err = ret;
		fields["domainname"] = ret.toString().replace("\"", "");

		ret = values.option("domain-name-servers");
		if (ret.isError()) err = ret;
		fields["dnsservers"] = ret;

		ret = values.option("routers");
		if (ret.isError()) err = ret;
		fields["gateway"] = ret;

		ret = values.name();
		if (ret.isError()) err = ret;
		fields["name"] = ret;
	}

	if (err.isError())
	{
		QMessageBox msg(QMessageBox::Warning, "Error", "There was an error reading settings: " + ret.errString(), QMessageBox::Ok);
		QAbstractButton *editbutton = msg.addButton("Edit Anyway", QMessageBox::RejectRole);
		msg.exec();
		if (msg.clickedButton() == editbutton)
		{
			if (QMessageBox::warning(this, "Warning", "Editing in this state may be dangerous. Continue?", QMessageBox::Yes | QMessageBox::No, QMessageBox::No) == QMessageBox::Yes)
			{
				return(fields);
			}
		}
		return(err);
	}

	return(fields);
}

#define SET_FIELD(field,function) if (name == field) \
	{ \
		ret = function; \
		if (ret.isError()) err = ret; \
	}
#define SET_VALUE_STRING(field,key) \
	if(value.toString() == "") \
	{\
		ret = values.removeValue(key); \
		if (ret.isError()) err = ret; \
	}\
	else SET_FIELD(field,values.setValue(key, value.toString()));
#define SET_VALUE_STRING_QUOTED(key,function) \
	SET_FIELD(field,values.setValue(key, "\"" + value.toString() + "\""));
#define SET_VALUE_BOOL(key,function) \
	SET_FIELD(field,values.setValue(key, value.toBool()));
#define SET_OPTION_STRING(field,key) \
	if(value.toString() == "") \
	{\
		ret = values.removeValue(key); \
		if (ret.isError()) err = ret; \
	}\
	else SET_FIELD(field,values.setOption(key, value.toString()));
#define SET_OPTION_STRING_QUOTED(key,function) \
	SET_FIELD(field,values.setOption(key, "\"" + value.toString() + "\""));
#define SET_OPTION_BOOL(key,function) \
	SET_FIELD(field,values.setOption(key, value.toBool()));

bool DhcpHostEditWidget::setData(QString id, QVariantMap fields)
{
	ReturnValue err = true;
	ReturnValue ret;

	ret = _manager->values(id);
	if (ret.isError()) err = ret;
	RdsDhcpValues values = ret;

	foreach(QString name, fields.keys())
	{
		QVariant value = fields[name];

		if (name == "domainname")
		{
			value = "\"" + value.toString() + "\"";
		}

		SET_OPTION_STRING("domainname", "domain-name");
		SET_OPTION_STRING("gateway", "routers");
		SET_OPTION_STRING("dnsservers", "domain-name-servers");
		SET_VALUE_STRING("ip", "fixed-address");
		SET_VALUE_STRING("mac", "hardware ethernet");

		if (name == "name")
		{
			//qDebug() << "setting name" << value;
			ret = values.setName(value.toString());
			if (ret.isError())
			{
				err = ret;
			}
			else
			{
				QString name = input();
				name = name.left(name.lastIndexOf("/") + 2);
				name += value.toString();

				QStringList outlist = outputs();
				for (int i = 0; i < outlist.size(); i++)
				{
					if (outlist[i] == input()) outlist[i] = name;
				}

				setInput(name);
				setOutputs(outlist);

				manager()->renameEntity(id, name);
			}
		}
	}

	if (err.isError())
	{
		qWarning() << "Failed to save:" << id << err.errString();
		return(false);
	}
	else return(true);
}

void DhcpHostEditWidget::entityUpdated(QString id)
{
	if (id != input()) return;

	if (unsavedChanges())
	{
		if (QMessageBox::question(this, "Group Changed",
		                          "Another user has made changes to the item you are editing. Would you like to overwrite your changes?",
		                          QMessageBox::Yes | QMessageBox::No, QMessageBox::No) == QMessageBox::Yes)
		{
			RdsEntityWidget::entityUpdated(id);
		}
	}
	else
	{
		RdsEntityWidget::entityUpdated(id);
	}
}

bool DhcpHostEditWidget::endSetData()
{
	ReturnValue ret = _manager->save();
	if (ret.isError())
	{
		QMessageBox::critical(this, "Error", "Failed to save changes: " + ret.errString());
		return(false);
	}

	RdsDaemonManager mgr;
	ret = mgr.init();

	if (ret.isError())
	{
		QMessageBox::critical(this, "Error", "Failed to get RdsDaemonManager service: " + ret.errString());
		return(false);
	}

	ret = mgr.restartService("Dhcp");

	if (ret.isError())
	{
		QMessageBox::critical(this, "Error", "Failed to restart the DHCP server: " + ret.errString());
		return(false);
	}

	return(true);
}

bool DhcpHostEditWidget::validate()
{
	if ((Gateway->text() != "") && !QRegExp(IP_VALIDATOR_STRONG).exactMatch(Gateway->text()))
	{
		QMessageBox::critical(this, "Error", "The gateway must be a valid IP address");
		return(false);
	}

	if ((Ip->text() != "") && !QRegExp(IP_VALIDATOR_STRONG).exactMatch(Ip->text()))
	{
		QMessageBox::critical(this, "Error", "The IP address must be a valid IP address");
		return(false);
	}

	return(true);
}


